# Monocyte - Monocyte - Search and Destroy unwanted AWS Resources relentlessly.
# Copyright 2015 Immobilien Scout GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import warnings
import logging
import boto3


class Resource(object):
    def __init__(self, resource, resource_type, resource_id, creation_date,
                 region=None, reason=None):
        self.wrapped = resource
        self.region = region
        self.resource_type = resource_type
        self.resource_id = resource_id
        self.creation_date = creation_date
        self.reason = reason

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return ("{class_name}({resource}, {resource_type}, {resource_id}, "
                "{creation_date}, region={region}, reason={reason})").format(
                class_name=self.__class__.__name__,
                resource=self.wrapped,
                resource_type=self.resource_type,
                resource_id=self.resource_id,
                creation_date=self.creation_date,
                region=self.region,
                reason=self.reason)

    def __repr__(self):
        return str(self)


HANDLER_PREFIX = "monocyte.handler."


class Handler(object):
    def __init__(self, region_filter, dry_run=True, logger=None, ignored_resources=None, whitelist=None):
        warnings.filterwarnings('error')
        self.region_filter = region_filter
        self.region_names = [region_name for region_name in self.fetch_region_names() if self.region_filter(region_name)]
        self.dry_run = dry_run
        self.ignored_resources = ignored_resources or []
        self.whitelist = whitelist or {}
        self.logger = logger or logging.getLogger(__name__)

    @property
    def resource_type(self):
        full_type = "%s.%s" % (self.__class__.__module__, self.__class__.__name__)
        return full_type.replace(HANDLER_PREFIX, "")

    @property
    def name(self):
        full_name = "%s.%s" % (self.__class__.__module__, self.__class__.__name__)
        return full_name.replace(HANDLER_PREFIX, "")

    def get_account_id(self):
        return boto3.client('sts').get_caller_identity().get('Account')

    def get_whitelist(self):
        return self.whitelist.get(self.get_account_id(), {})

    def fetch_region_names(self):
        raise NotImplementedError("Should have implemented this")

    def fetch_unwanted_resources(self):
        raise NotImplementedError("Should have implemented this")

    def to_string(self, resource):
        raise NotImplementedError("Should have implemented this")

    def delete(self, resource):
        raise NotImplementedError("Should have implemented this")



# Monocyte - Search and Destroy unwanted AWS Resources relentlessly.
# Copyright 2015 Immobilien Scout GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from boto import rds2
from monocyte.handler import Resource, Handler

SKIPPING_CREATION_STATEMENT = "Currently in creation. Skipping."
SKIPPING_AUTOGENERATED_STATEMENT = "Not a manually created Snapshot. Skipping."
SKIPPING_DELETION_STATEMENT = "Deletion already in progress. Skipping."
DELETION_STATEMENT = "Initiating deletion sequence for %s."

CREATION_STATUS = "creating"
AUTOMATED_STATUS = "automated"
DELETION_STATUS = "deleting"


class Instance(Handler):

    def fetch_region_names(self):
        return [region.name for region in rds2.regions()]

    def fetch_unwanted_resources(self):
        for region_name in self.region_names:
            connection = rds2.connect_to_region(region_name)
            resources = connection.describe_db_instances() or []
            for resource in resources["DescribeDBInstancesResponse"]["DescribeDBInstancesResult"]["DBInstances"]:
                resource_wrapper = Resource(resource=resource,
                                            resource_type=self.resource_type,
                                            resource_id=resource["DBInstanceIdentifier"],
                                            creation_date=resource["InstanceCreateTime"],
                                            region=region_name)
                if resource['DBInstanceIdentifier'] in self.ignored_resources:
                    self.logger.info('IGNORE ' + self.to_string(resource_wrapper))
                    continue
                yield resource_wrapper

    def to_string(self, resource):
        return "Database Instance found in {region}, ".format(**vars(resource)) + \
               "with name {DBInstanceIdentifier}, with status {DBInstanceStatus}".format(**resource.wrapped)

    def delete(self, resource):
        if self.dry_run:
            return
        if resource.wrapped["DBInstanceStatus"] == DELETION_STATUS:
            warnings.warn(Warning(SKIPPING_DELETION_STATEMENT))
        self.logger.info(DELETION_STATEMENT % resource.wrapped["DBInstanceIdentifier"])
        connection = rds2.connect_to_region(resource.region)
        connection.delete_db_instance(resource.wrapped["DBInstanceIdentifier"], skip_final_snapshot=True)


class Snapshot(Handler):

    def fetch_region_names(self):
        return [region.name for region in rds2.regions()]

    def fetch_unwanted_resources(self):
        for region_name in self.region_names:
            connection = rds2.connect_to_region(region_name)
            resources = connection.describe_db_snapshots() or []
            for resource in resources["DescribeDBSnapshotsResponse"]["DescribeDBSnapshotsResult"]["DBSnapshots"]:
                resource_wrapper = Resource(resource=resource,
                                            resource_type=self.resource_type,
                                            resource_id=resource["DBSnapshotIdentifier"],
                                            creation_date=resource["SnapshotCreateTime"],
                                            region=region_name)
                if resource['DBSnapshotIdentifier'] in self.ignored_resources:
                    self.logger.info('IGNORE ' + self.to_string(resource_wrapper))
                    continue
                yield resource_wrapper

    def to_string(self, resource):
        return "Database Snapshot found in {region}, ".format(**vars(resource)) + \
               "with name {DBSnapshotIdentifier}, with status {Status}".format(**resource.wrapped)

    def delete(self, resource):
        if self.dry_run:
            return
        if resource.wrapped["Status"] == DELETION_STATUS:
            warnings.warn(Warning(SKIPPING_DELETION_STATEMENT))
        if resource.wrapped["Status"] == CREATION_STATUS:
            warnings.warn(Warning(SKIPPING_CREATION_STATEMENT))
        if resource.wrapped["SnapshotType"] == AUTOMATED_STATUS:
            warnings.warn(Warning(SKIPPING_AUTOGENERATED_STATEMENT))

        self.logger.info(DELETION_STATEMENT % resource.wrapped["DBSnapshotIdentifier"])
        connection = rds2.connect_to_region(resource.region)
        connection.delete_db_snapshot(resource.wrapped["DBSnapshotIdentifier"])



from __future__ import print_function, absolute_import, division

import boto3
import json
import logging


class AwsSQSPlugin(object):

    def __init__(self, unwanted_resources, problematic_resources, dry_run,
                 queue_account=None, queue_name=None, queue_region=None,):
        self.queue_account = queue_account
        self.queue_name = queue_name
        self.queue_region = queue_region
        self.unwanted_resources = unwanted_resources
        self.problematic_resources = problematic_resources
        self.dry_run = dry_run

        self.logger = logging.getLogger(__name__)

    def _get_account_alias(self):
        iam = boto3.client('iam')
        response = iam.list_account_aliases()
        return response['AccountAliases'][0]

    def monocyte_status(self):
        if self.unwanted_resources or self.problematic_resources:
            return "Found {0} unwanted and {1} problematic resources.".format(
                len(self.unwanted_resources), len(self.problematic_resources)
            )
        return "no issues"

    def get_body(self):
        body = {
            'status': self.monocyte_status(),
            'account': self._get_account_alias()
        }
        return json.dumps(body)

    def send_message(self, body):
        sqs = boto3.client('sqs', region_name=self.queue_region)
        response = sqs.get_queue_url(QueueName=self.queue_name, QueueOwnerAWSAccountId=self.queue_account)
        sqs.send_message(QueueUrl=response['QueueUrl'], MessageBody=body)

    def run(self):
        try:
            self.send_message(self.get_body())
        except Exception as exc:
            self.logger.error("%s.run() failed to send to SQS: %s",
                              self.__class__.__name__, exc)



"""Check for ACM certificates that will expire soon

ACM is the "AWS Certificate Manager" that usually renews your certificates
automatically, so you don't have to. However, there's a few details in the
fine print at

    https://docs.aws.amazon.com/acm/latest/userguide/acm-renewal.html

which means it is not 100% guaranteed to work. And once we are at it, we
can also check imported certificates (which cannot be renewed by ACM).
"""
from __future__ import absolute_import, print_function, division

import datetime
import boto3
from monocyte.handler import Resource, Handler

# ACM attempts to renew SSL certificates 60 before expiration. If it
# is still not renewed 55 days before expiration, something is wrong.
MIN_VALID_DAYS = 55

class Certificate(Handler):
    def fetch_region_names(self):
        # Since we want to check all regions, regardless of what's allowed or
        # not, we handle multi-region stuff ourselves.
        return []

    def fetch_unwanted_resources(self):
        session = boto3.session.Session()
        region_names = session.get_available_regions('acm')
        unwanted_resources = []

        for region_name in region_names:
            unwanted_resources.extend(list(self._fetch_unwanted_resources(region_name)))
        return unwanted_resources

    def _fetch_unwanted_resources(self, region_name):
        client = boto3.client('acm', region_name=region_name)
        response = client.list_certificates(CertificateStatuses=['ISSUED'])
        certificate_arns = [summary['CertificateArn'] for summary in response['CertificateSummaryList']]

        limit = datetime.datetime.now() + datetime.timedelta(days=MIN_VALID_DAYS)

        for certificate_arn in certificate_arns:
            response = client.describe_certificate(CertificateArn=certificate_arn)
            certificate = response['Certificate']

            # Remove time zone information so we can compare with normal datetimes.
            not_after = datetime.datetime.replace(certificate['NotAfter'], tzinfo=None)

            if not_after > limit:
                continue

            resource_wrapper = Resource(
                resource="Certificate for " + certificate['DomainName'],
                resource_type=self.resource_type,
                resource_id=certificate_arn,
                creation_date=certificate.get('CreatedAt', certificate.get('ImportedAt')),
                region="global",
                reason="will expired soon")
            yield resource_wrapper

    def to_string(self, resource):
        return "{summary} with ARN {arn}, created at {creation_date} will expire soon.".format(
            summary=resource.wrapped, arn=resource.resource_id,
            creation_date=resource.creation_date)

    def delete(self, resource):
        if self.dry_run:
            return
        raise NotImplementedError("ACM certificates are never deleted.")



# Monocyte - Search and Destroy unwanted AWS Resources relentlessly.
# Copyright 2015 Immobilien Scout GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from boto import dynamodb2
from monocyte.handler import Resource, Handler


class Table(Handler):
    def fetch_region_names(self):
        return [region.name for region in dynamodb2.regions()]

    def fetch_unwanted_resources(self):
        for region_name in self.region_names:
            connection = dynamodb2.connect_to_region(region_name)
            names = connection.list_tables(limit=100) or {}
            for name in names.get("TableNames"):
                resource = connection.describe_table(name)
                resource_wrapper = Resource(resource=resource["Table"],
                                            resource_type=self.resource_type,
                                            resource_id=resource["Table"]["TableName"],
                                            creation_date=resource["Table"]["CreationDateTime"],
                                            region=region_name)
                if name in self.ignored_resources:
                    self.logger.info('IGNORE ' + self.to_string(resource_wrapper))
                    continue

                yield resource_wrapper

    def to_string(self, resource):
        table = resource.wrapped
        return "DynamoDB Table found in {0}, ".format(resource.region) + \
               "with name {0}, created {1}, with state {2}".format(
                   table["TableName"],
                   datetime.datetime.fromtimestamp(table["CreationDateTime"]).strftime('%Y-%m-%d %H:%M:%S.%f'),
                   table["TableStatus"])

    def delete(self, resource):
        if self.dry_run:
            return
        connection = dynamodb2.connect_to_region(resource.region)
        self.logger.info("Initiating deletion sequence for {0}.".format(resource.wrapped["TableName"]))
        connection.delete_table(resource.wrapped["TableName"])



# Monocyte - Search and Destroy unwanted AWS Resources relentlessly.
# Copyright 2015 Immobilien Scout GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import boto3
from monocyte.handler import Resource, Handler

US_STANDARD_REGION = "us-east-1"
SIGV4_REGIONS = ['eu-central-1']
AVAILABILITY_ZONES = {'EU': 'eu-west-1', None: US_STANDARD_REGION}


class Bucket(Handler):
    def map_location(self, region):
        return AVAILABILITY_ZONES.get(region, region)

    def get_client(self):
        return boto3.client('s3', region_name='eu-central-1')

    def fetch_region_names(self):
        session = boto3.session.Session()
        return session.get_available_regions('s3')

    def fetch_unwanted_resources(self):
        client = self.get_client()
        response = client.list_buckets()
        buckets = [(bucket['Name'], bucket['CreationDate'])
                   for bucket in response['Buckets']]

        for bucket_name, creation_date in buckets:
            try:
                response = client.get_bucket_location(Bucket=bucket_name)
            except Exception:
                # This happens when a bucket was deleted shortly after we
                # found it. E.g. during concurrent integration tests.
                self.logger.exception("Failed to get location for bucket %r:",
                                      bucket_name)
                continue
            region_name = self.map_location(response['LocationConstraint'])
            if region_name not in self.region_names or self.is_on_whitelist(bucket_name):
                self.logger.debug("Bucket %s in region %s is OK.",
                                  bucket_name, region_name)
                continue

            self.logger.info("Reporting bucket %s in region %s as unwanted.",
                             bucket_name, region_name)
            resource_wrapper = Resource(
                resource="Bucket " + bucket_name,
                resource_type=self.resource_type,
                resource_id=bucket_name,
                creation_date=creation_date,
                region=region_name)
            yield resource_wrapper

    def to_string(self, resource):
        return "s3 bucket found in {0}, with name {1}, created {2}".format(
            resource.region, resource.resource_id, resource.creation_date)

    def is_on_whitelist(self, bucket_name):
        bucket_arn = "arn:aws:s3:::%s" % bucket_name
        whitelist_arns = self.get_whitelist().get('Arns', [])
        for arn_with_reason in whitelist_arns:
            if bucket_arn == arn_with_reason['Arn']:
                return True

        return False

    def delete(self, resource):
        if self.dry_run:
            return

        if isinstance(resource, Resource):
            bucket_name = resource.resource_id
        else:
            bucket_name = resource
        client = self.get_client()

        while True:
            response = client.list_objects_v2(Bucket=bucket_name)
            if response['KeyCount'] == 0:
                break
            keys = [item['Key'] for item in response['Contents']]
            delete_spec = {'Objects': [{'Key': key} for key in keys]}
            client.delete_objects(Bucket=bucket_name, Delete=delete_spec)
        try:
            client.delete_bucket(Bucket=bucket_name)
        except Exception:
            self.logger.exception("Failed to delete bucket %r:" % resource)
            raise



from __future__ import print_function, absolute_import, division

import boto3
import json

from .ses_plugin import AwsSesPlugin


class StatusMailPlugin(AwsSesPlugin):
    def __init__(self, unwanted_resources, problematic_resources, dry_run, **kwargs):
        self.subject = 'AWS Compliance Checker - Your action is required'
        super(StatusMailPlugin, self).__init__(unwanted_resources, problematic_resources, dry_run, subject=self.subject, **kwargs)

    @property
    def body(self):
        if self.dry_run:
            unwanted_resources_info = "Please check the following resources:"
        else:
            unwanted_resources_info = "Please check the following deleted resources:"
        email_body = '''Dear AWS User,

our Compliance checker found some issues in your account.
{0}

Account: {1}\n'''.format(unwanted_resources_info, self._get_account_alias())

        email_body += self._handle_resources(self.unwanted_resources)

        if self.problematic_resources:
            email_body += ("\nAdditionally we had issues checking the following "
                           "resource, please ensure that they are in the proper region:\n")
            email_body += self._handle_resources(self.problematic_resources)

        email_footer = '\n Kind regards.\n\tYour Compliance Team'
        email_body += email_footer

        return email_body

    def _handle_resources(self, resources):
        return_text = ""
        regions = sorted(list(set([res.region for res in resources])))
        res_types = sorted(list(set([res.resource_type for res in resources])))

        for region in regions:
            region_text = "Region: {0}\n".format(region)
            return_text += region_text
            for res_type in res_types:
                selected_res = (resource for resource in resources
                                if resource.region == region and resource.resource_type == res_type)
                for resource in selected_res:
                    res_text = "\t{0} with identifier {1}, created {2}.".format(
                        res_type, resource.resource_id, resource.creation_date)
                    if resource.reason:
                        res_text += ' ' + resource.reason
                    res_text += '\n'
                    return_text += res_text
        return return_text or "\tNone\n"

    def _get_account_alias(self):
        iam = boto3.client('iam')
        response = iam.list_account_aliases()
        return response['AccountAliases'][0]

    def run(self):
        if self.unwanted_resources or self.problematic_resources:
            self.send_email()


class UsofaStatusMailPlugin(StatusMailPlugin):
    """StatusMailPlugin that finds additional recipients via usofa"""
    def __init__(self, unwanted_resources, problematic_resources, dry_run, usofa_bucket_name=None, **kwargs):
        super(UsofaStatusMailPlugin, self).__init__(unwanted_resources, problematic_resources, dry_run, **kwargs)
        self.usofa_bucket_name = usofa_bucket_name

    def _get_usofa_data(self):
        s3 = boto3.client('s3', region_name=self.region)
        response = s3.get_object(Bucket=self.usofa_bucket_name, Key='accounts.json')
        account_data = json.loads(response['Body'].read().decode('utf-8'))
        return account_data

    @property
    def recipients(self):
        usofa = self._get_usofa_data()
        account_alias = self._get_account_alias()
        responsible = usofa[account_alias]['email']

        if self.mail_recipients:
            recipients = list(self.mail_recipients)
            recipients.append(responsible)
            return recipients
        return [responsible]
